from asyncio import TimeoutError
from datetime import datetime, timedelta
from typing import Awaitable, Collection, Dict, Mapping, Optional, Union, cast

from aiohttp import (
    ClientError,
    ClientSession,
    ClientTimeout,
    ServerTimeoutError,
    TCPConnector,
)
from grpc import RpcError, Status
from typing_extensions import Literal

from ..._deadline import monotonic_time_from_deadline
from ..._typing import assert_unreachable
from ...options import AuthorizerOptions
from ...identity import Identity
from ...resource_context import ResourceContext
from ._protocol import AuthorizerClientProtocol, DecisionTree


class AuthorizerRestClient(AuthorizerClientProtocol):
    def __init__(
        self,
        *,
        identity: Identity,
        options: AuthorizerOptions,
    ):
        self._options = options

        self._identity_context_field: Mapping[str, str] = {
            "type": identity.type_field,
        }

        if identity.identity_field is not None:
            self._identity_context_field["identity"] = identity.identity_field

    @property
    def _authorizer_api_url_base(self) -> str:
        return f"{self._options.url}/api/v1/authz"

    @property
    def _headers(self) -> Mapping[str, str]:
        headers = {"Content-Type": "application/json", **self._options.auth_headers}

        return headers

    def _authorizer_session(self, deadline: Optional[Union[datetime, timedelta]]) -> ClientSession:
        connector = None
        if self._options.ssl_context is not None:
            connector = TCPConnector(ssl_context=self._options.ssl_context)

        return ClientSession(
            connector=connector,
            timeout=ClientTimeout(
                total=(monotonic_time_from_deadline(deadline) if deadline is not None else None),
            ),
        )

    @staticmethod
    def _policy_path_separator_field(
        policy_path_separator: Literal["DOT", "SLASH"],
    ) -> Literal["PATH_SEPARATOR_DOT", "PATH_SEPARATOR_SLASH"]:
        if policy_path_separator == "DOT":
            return "PATH_SEPARATOR_DOT"
        elif policy_path_separator == "SLASH":
            return "PATH_SEPARATOR_SLASH"
        else:
            assert_unreachable(policy_path_separator)

    async def decision_tree(
        self,
        *,
        policy_path_root: str,
        decisions: Collection[str],
        policy_instance_name: Optional[str] = None,
        policy_instance_label: Optional[str] = None,
        resource_context: Optional[ResourceContext] = None,
        policy_path_separator: Optional[Literal["DOT", "SLASH"]] = None,
        deadline: Optional[Union[datetime, timedelta]] = None,
    ) -> DecisionTree:
        options = {}
        if policy_path_separator is not None:
            options["pathSeparator"] = self._policy_path_separator_field(policy_path_separator)

        body = {
            "policyContext": {
                "path": policy_path_root,
                "decisions": tuple(decisions),
            },
            "identityContext": self._identity_context_field,
            "resourceContext": resource_context,
            "options": options,
            "policyInstance": {
                "name": policy_instance_name,
                "instanceLabel": policy_instance_label,
            },
        }

        async with self._authorizer_session(deadline=deadline) as session:
            url = f"{self._authorizer_api_url_base}/decisiontree"

            try:
                async with session.post(url, headers=self._headers, json=body) as response:
                    response_json = await cast(Awaitable[object], response.json())
            except ServerTimeoutError as error:
                # ServerTimeoutError is a TimeoutError but we only want to expose the latter
                raise TimeoutError(*error.args)  # type: ignore[misc]
            except ClientError as error:
                raise ConnectionError(*error.args) from error  # type: ignore[misc]

        return self._validate_decision_tree_response(response_json)

    @staticmethod
    def _raise_if_server_error(response: Mapping[object, object]) -> None:
        if response.keys() == {"code", "message", "details"}:
            raise RpcError(
                Status(response["code"]),
                (str(response["message"]) if response["message"] is not None else None),
                response["details"],
            )

    @classmethod
    def _validate_decision_tree_response(cls, response: object) -> DecisionTree:
        error = TypeError("Received unexpected response data", response)

        if not isinstance(response, dict):
            raise error

        if "path" not in response:
            cls._raise_if_server_error(response)
            raise error

        tree = response["path"]

        if not isinstance(tree, dict):
            raise error

        for path, decisions in tree.items():
            if not isinstance(path, str):
                raise error

            if not isinstance(decisions, dict):
                raise error

            for name, decision in decisions.items():
                if not isinstance(name, str):
                    raise error

                if not isinstance(decision, bool):
                    raise error

        return cast(DecisionTree, tree)

    async def decisions(
        self,
        *,
        policy_path: str,
        decisions: Collection[str],
        policy_instance_name: Optional[str] = None,
        policy_instance_label: Optional[str] = None,
        resource_context: Optional[ResourceContext] = None,
        deadline: Optional[Union[datetime, timedelta]] = None,
    ) -> Dict[str, bool]:
        body = {
            "policyContext": {
                "path": policy_path,
                "decisions": list(decisions),
            },
            "identityContext": self._identity_context_field,
            "resourceContext": resource_context,
            "policyInstance": {
                "name": policy_instance_name,
                "instanceLabel": policy_instance_label,
            },
        }

        async with self._authorizer_session(deadline=deadline) as session:
            url = f"{self._authorizer_api_url_base}/is"

            try:
                async with session.post(url, headers=self._headers, json=body) as response:
                    response_json = await cast(Awaitable[object], response.json())
            except ServerTimeoutError as error:
                # ServerTimeoutError is a TimeoutError but we only want to expose the latter
                raise TimeoutError(*error.args)  # type: ignore[misc]
            except ClientError as error:
                raise ConnectionError(*error.args) from error  # type: ignore[misc]

        return self._validate_decision_response(response_json)

    @classmethod
    def _validate_decision_response(cls, response: object) -> Dict[str, bool]:
        error = TypeError("Received unexpected response data")

        if not isinstance(response, dict):
            raise error

        if "decisions" not in response:
            cls._raise_if_server_error(response)
            raise error

        decisions = response["decisions"]

        if not isinstance(decisions, list):
            raise error

        result = {}
        for decision_json in decisions:
            if not isinstance(decision_json, dict):
                raise error

            if "decision" not in decision_json:
                raise error

            if not isinstance(decision_json["decision"], str):
                raise error

            if not isinstance(decision_json["is"], bool):
                raise error

            result[decision_json["decision"]] = decision_json["is"]

        return result



from datetime import datetime, timedelta
from typing import Collection, Dict, Optional, Union

from typing_extensions import Literal, Protocol

from ...resource_context import ResourceContext

__all__ = [
    "AuthorizerClientProtocol",
    "DecisionTree",
]


DecisionTree = Dict[str, Dict[str, bool]]


class AuthorizerClientProtocol(Protocol):
    async def decision_tree(
        self,
        *,
        policy_path_root: str,
        decisions: Collection[str],
        policy_instance_name: Optional[str] = None,
        policy_instance_label: Optional[str] = None,
        resource_context: Optional[ResourceContext] = None,
        policy_path_separator: Optional[Literal["DOT", "SLASH"]] = None,
        deadline: Optional[Union[datetime, timedelta]] = None,
    ) -> DecisionTree:
        ...

    async def decisions(
        self,
        *,
        policy_path: str,
        decisions: Collection[str],
        policy_instance_name: Optional[str] = None,
        policy_instance_label: Optional[str] = None,
        resource_context: Optional[ResourceContext] = None,
        deadline: Optional[Union[datetime, timedelta]] = None,
    ) -> Dict[str, bool]:
        ...



from datetime import datetime, timedelta
from typing import Collection, Dict, Optional, Union

from typing_extensions import Literal

from ..._typing import assert_unreachable
from ...options import AuthorizerOptions
from ...identity import Identity
from ...resource_context import ResourceContext
from ._protocol import AuthorizerClientProtocol, DecisionTree
from .grpc import AuthorizerGrpcClient
from .rest import AuthorizerRestClient

__all__ = ["AuthorizerClient", "DecisionTree"]


class AuthorizerClient(AuthorizerClientProtocol):
    def __init__(
        self,
        *,
        identity: Identity,
        options: AuthorizerOptions,
    ):
        service_type = options.service_type
        if service_type == "gRPC":
            self._client: AuthorizerClientProtocol = AuthorizerGrpcClient(
                identity=identity,
                options=options,
            )
        elif service_type == "REST":
            self._client = AuthorizerRestClient(
                identity=identity,
                options=options,
            )
        else:
            assert_unreachable(service_type)

    async def decision_tree(
        self,
        *,
        policy_path_root: str,
        decisions: Collection[str],
        policy_instance_name: Optional[str] = None,
        policy_instance_label: Optional[str] = None,
        resource_context: Optional[ResourceContext] = None,
        policy_path_separator: Optional[Literal["DOT", "SLASH"]] = None,
        deadline: Optional[Union[datetime, timedelta]] = None,
    ) -> DecisionTree:
        return await self._client.decision_tree(
            policy_path_root=policy_path_root,
            decisions=decisions,
            policy_instance_name=policy_instance_name,
            policy_instance_label=policy_instance_label,
            resource_context=resource_context,
            policy_path_separator=policy_path_separator,
            deadline=deadline,
        )

    async def decisions(
        self,
        *,
        policy_path: str,
        decisions: Collection[str],
        policy_instance_name: Optional[str] = None,
        policy_instance_label: Optional[str] = None,
        resource_context: Optional[ResourceContext] = None,
        deadline: Optional[Union[datetime, timedelta]] = None,
    ) -> Dict[str, bool]:
        return await self._client.decisions(
            policy_path=policy_path,
            decisions=decisions,
            policy_instance_name=policy_instance_name,
            policy_instance_label=policy_instance_label,
            resource_context=resource_context,
            deadline=deadline,
        )



from typing import Dict, Optional, cast

from typing_extensions import Literal, overload

from ._typing import assert_unreachable

__all__ = ["Identity", "IdentityType"]


IdentityType = Literal[
    "NONE",
    "SUBJECT",
    "JWT",
]

IdentityTypeField = Literal[
    "IDENTITY_TYPE_NONE",
    "IDENTITY_TYPE_SUB",
    "IDENTITY_TYPE_JWT",
    "IDENTITY_TYPE_UNKNOWN",
]


class Identity:
    @overload
    def __init__(self, *, type: Literal["NONE"]):
        ...

    @overload
    def __init__(self, *, type: Literal["SUBJECT"], subject: str):
        ...

    @overload
    def __init__(self, *, type: Literal["JWT"], token: str):
        ...

    def __init__(
        self,
        *,
        type: IdentityType,
        subject: Optional[str] = None,
        token: Optional[str] = None,
    ):
        self._type = type

        if self._type == "NONE":
            self._identity = None
        elif self._type == "SUBJECT":
            self._identity = subject
        elif self._type == "JWT":
            self._identity = token
        else:
            assert_unreachable(self._type)

    @property
    def type(self) -> IdentityType:
        return self._type

    @property
    def type_field(self) -> IdentityTypeField:
        if self._type == "NONE":
            return "IDENTITY_TYPE_NONE"
        elif self._type == "SUBJECT":
            return "IDENTITY_TYPE_SUB"
        elif self._type == "JWT":
            return "IDENTITY_TYPE_JWT"
        else:
            assert_unreachable(self._type)

    @property
    def identity_field(self) -> Optional[str]:
        return self._identity

    def __repr__(self) -> str:
        fields: Dict[str, str] = {"type": self._type}

        if self._type == "NONE":
            pass
        elif self._type == "SUBJECT":
            fields["subject"] = cast(str, self._identity)
        elif self._type == "JWT":
            fields["token"] = cast(str, self._identity)
        else:
            assert_unreachable(self._type)

        fields_string = ", ".join(f"{name}={value}" for name, value in fields.items())
        return f"IdentityContext({fields_string})"



import json
from contextlib import asynccontextmanager
from datetime import datetime, timedelta
from typing import AsyncGenerator, Collection, Dict, Mapping, Optional, Union
from urllib.parse import urlparse

import google.protobuf.struct_pb2 as structpb
from aserto.authorizer.v2.api import (
    IdentityContext,
    PolicyContext,
    IdentityType,
    PolicyInstance,
)
from aserto.authorizer.v2 import (
    AuthorizerStub,
    DecisionTreeOptions,
    DecisionTreeRequest,
    DecisionTreeResponse,
    PathSeparator,
    IsRequest,
)

import grpc
import grpc.aio as grpcaio
from typing_extensions import Literal

from ..._deadline import monotonic_time_from_deadline
from ..._typing import assert_unreachable
from ...options import AuthorizerOptions
from ...identity import Identity
from ...resource_context import ResourceContext
from ._protocol import AuthorizerClientProtocol, DecisionTree


class AuthorizerGrpcClient(AuthorizerClientProtocol):
    def __init__(
        self,
        *,
        tenant_id: Optional[str] = None,
        identity: Identity,
        options: AuthorizerOptions,
    ):
        self._tenant_id = tenant_id
        self._options = options
        self._identity_context_field = IdentityContext(
            identity=identity.identity_field or "",
            type=IdentityType.Value(identity.type_field),
        )

    @property
    def _headers(self) -> Mapping[str, str]:
        return self._options.auth_headers

    @property
    def _metadata(self) -> grpcaio.Metadata:
        return grpcaio.Metadata(*tuple(self._headers.items()))

    @asynccontextmanager  # type: ignore[misc]
    async def _authorizer_client(self, deadline: Optional[Union[datetime, timedelta]]) -> AsyncGenerator[AuthorizerStub, None]:  # type: ignore[misc]
        result = urlparse(self._options.url)
        channel = grpcaio.secure_channel(
            target=f"{result.hostname}:{result.port}",
            credentials=grpc.ssl_channel_credentials(self._options.cert),
        )

        async with channel as channel:
            yield AuthorizerStub(channel)

    @staticmethod
    def _policy_path_separator_field(
        policy_path_separator: Literal["DOT", "SLASH"]
    ) -> PathSeparator.ValueType:
        if policy_path_separator == "DOT":
            return PathSeparator.PATH_SEPARATOR_DOT
        elif policy_path_separator == "SLASH":
            return PathSeparator.PATH_SEPARATOR_SLASH
        else:
            assert_unreachable(policy_path_separator)

    @classmethod
    def _serialize_resource_context(cls, resource_context: object) -> structpb.Struct:
        try:
            json.dumps(resource_context)
        except ValueError as error:
            if error.args == ("Circular reference detected",):  # type: ignore[misc]
                raise TypeError("Resource context is circularly defined")
            else:
                raise TypeError("Invalid resource context")

        proto_value = cls._serialize_resource_context_value(resource_context)
        return proto_value.struct_value

    @classmethod
    def _serialize_resource_context_value(cls, resource_value: object) -> structpb.Value:
        # `Mapping` is a subclass of `Collection` so this check must come first
        if isinstance(resource_value, Mapping):
            struct_value = structpb.Struct()
            struct_value.update(resource_value)
            return structpb.Value(struct_value=struct_value)
        # `str` is a subclass of `Collection` so this check must come first
        elif isinstance(resource_value, str):
            return structpb.Value(string_value=resource_value)
        elif isinstance(resource_value, Collection):
            list_value = structpb.ListValue()
            for value in resource_value:
                list_value.append(cls._serialize_resource_context_value(value))
            return structpb.Value(list_value=list_value)
        # `bool` is subclass of `int` so this check must come first
        elif isinstance(resource_value, bool):
            return structpb.Value(bool_value=resource_value)
        elif isinstance(resource_value, (int, float)):
            return structpb.Value(number_value=float(resource_value))
        elif resource_value is None:
            return structpb.Value(null_value=structpb.NullValue.NULL_VALUE)
        else:
            raise TypeError("Invalid resource context")

    async def decision_tree(
        self,
        *,
        policy_path_root: str,
        decisions: Collection[str],
        policy_instance_name: Optional[str] = None,
        policy_instance_label: Optional[str] = None,
        resource_context: Optional[ResourceContext] = None,
        policy_path_separator: Optional[Literal["DOT", "SLASH"]] = None,
        deadline: Optional[Union[datetime, timedelta]] = None,
    ) -> DecisionTree:
        options = DecisionTreeOptions()
        if policy_path_separator is not None:
            options.path_separator = self._policy_path_separator_field(policy_path_separator)

        try:
            async with self._authorizer_client(deadline=deadline) as client:
                response = await client.DecisionTree(
                    DecisionTreeRequest(
                        policy_context=PolicyContext(
                            path=policy_path_root,
                            decisions=list(decisions),
                        ),
                        identity_context=self._identity_context_field,
                        resource_context=self._serialize_resource_context(resource_context or {}),
                        options=options,
                        policy_instance=PolicyInstance(
                            name=policy_instance_name,
                            instance_label=policy_instance_label,
                        ),
                    ),
                    metadata=self._metadata,
                    timeout=(monotonic_time_from_deadline(deadline) if deadline is not None else None),
                )
        except (OSError, grpc.RpcError) as error:
            raise ConnectionError(*error.args) from error  # type: ignore[misc]

        return self._validate_decision_tree(response)

    @staticmethod
    def _validate_decision_tree(response: DecisionTreeResponse) -> DecisionTree:
        error = TypeError("Received unexpected response data")

        decision_tree: DecisionTree = {}

        for path, decisions in response.path.fields.items():
            if decisions.WhichOneof("kind") != "struct_value":
                raise error

            for name, decision in decisions.struct_value.fields.items():
                if decision.WhichOneof("kind") != "bool_value":
                    raise error

                decision_tree.setdefault(path, {})[name] = decision.bool_value

        return decision_tree

    async def decisions(
        self,
        *,
        decisions: Collection[str],
        policy_name: str,
        policy_path: str,
        resource_context: Optional[ResourceContext] = None,
        deadline: Optional[Union[datetime, timedelta]] = None,
    ) -> Dict[str, bool]:
        try:
            async with self._authorizer_client(deadline=deadline) as client:
                response = await client.Is(
                    IsRequest(
                        policy_context=PolicyContext(
                            name=policy_name,
                            path=policy_path,
                            decisions=list(decisions),
                        ),
                        identity_context=self._identity_context_field,
                        resource_context=self._serialize_resource_context(resource_context or {}),
                    ),
                    metadata=self._metadata,
                    timeout=(monotonic_time_from_deadline(deadline) if deadline is not None else None),
                )
        except (OSError, grpc.RpcError) as error:
            raise ConnectionError(*error.args) from error  # type: ignore[misc]

        results = {}
        for decision_object in response.decisions:
            results[decision_object.decision] = getattr(decision_object, "is")

        return results



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['aserto', 'aserto.client', 'aserto.client.api', 'aserto.client.api.authorizer']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0',
 'aserto-authorizer==0.0.4',
 'grpcio>=1.49.1,<2.0.0',
 'protobuf>=4.21.7,<5.0.0',
 'typing-extensions>=3.10.0,<4.0.0']

setup_kwargs = {
    'name': 'aserto',
    'version': '0.3.0rc1',
    'description': 'Aserto API client',
    'long_description': '# Aserto API client\nHigh-level client interface to Aserto\'s APIs.\n\nAt the moment this only supports interacting with Aserto\'s [Authorizer service](https://docs.aserto.com/docs/authorizer-guide/overview).\n## Installation\n### Using Pip\n```sh\npip install aserto\n```\n### Using Poetry\n```sh\npoetry add aserto\n```\n## Usage\n```py\nfrom aserto.client import AuthorizerOptions, Identity\nfrom aserto.client.api.authorizer import AuthorizerClient\n\n\nclient = AuthorizerClient(\n    identity=Identity(type="NONE"),\n    options=AuthorizerOptions(\n        api_key=ASERTO_API_KEY,\n        tenant_id=ASERTO_TENANT_ID,\n        service_type="gRPC",\n    ),\n)\n\nresult = await client.decision_tree(\n    decisions=["visible", "enabled", "allowed"],\n    policy_name=ASERTO_POLICY_NAME,\n    policy_path_root=ASERTO_POLICY_PATH_ROOT,\n    policy_path_separator="DOT",\n)\n\nassert result == {\n    "GET.your.policy.path": {\n        "visible": True,\n        "enabled": True,\n        "allowed": False,\n    },\n}\n```\n',
    'author': 'Aserto, Inc.',
    'author_email': 'pypi@aserto.com',
    'maintainer': 'authereal',
    'maintainer_email': 'authereal@aserto.com',
    'url': 'https://github.com/aserto-dev/aserto-python/tree/HEAD/packages/aserto',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
